from __future__ import annotations
import attrs
import functools
import logging
import numpy as np
import pyvista as pv
from skspatial.objects import Line, Plane, Vector
import types
import typing as tp

from NaviNIBS.Navigator.Model import Session
from NaviNIBS.Navigator.Model.Calculations import calculateAngleFromMidlineFromCoilToMRITransf, getClosestPointToPointOnMesh
from NaviNIBS.Navigator.Model.Samples import Sample
from NaviNIBS.Navigator.Model.Targets import Target
from NaviNIBS.util.pyvista.dataset import find_closest_point
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.Transforms import applyTransform, composeTransform, invertTransform, estimateAligningTransform


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


T = tp.TypeVar('T')


@attrs.define(frozen=True)
class MetricSpecification:
    getter: tp.Callable[[], float]
    units: str
    label: str
    doShowByDefault: bool = True


@attrs.define
class PoseMetricCalculator:
    _session: Session = attrs.field(repr=False)
    _sample: tp.Optional[Sample]

    _cachedValues: dict[str, tp.Any] = attrs.field(init=False, factory=dict, repr=False)
    _supportedMetrics: list[MetricSpecification] = attrs.field(init=False, factory=list)

    sigCacheReset: Signal = attrs.field(init=False, factory=Signal)
    """
    Emitted whenever previously cached values are cleared, e.g. due to change in sample orientation, sample, or target.
    """

    def __attrs_post_init__(self):
        # self.session.MNIRegistration.sigTransformChanged.connect(lambda *args: self._clearCachedValues())  # TODO: debug, uncomment
        self.session.headModel.sigDataChanged.connect(lambda *args: self._clearCachedValues())
        self.session.targets.sigItemsChanged.connect(self._onTargetsChanged)
        self.session.subjectRegistration.fiducials.sigItemsChanged.connect(self._onFiducialsChanged)
        if self._sample is not None:
            self._sample.sigItemChanged.connect(self._onSampleChanged)

        self._supportedMetrics.extend([
            MetricSpecification(getter=self.getTargetErrorInBrain, units=' mm', label='Target error in brain'),
            MetricSpecification(getter=self.getTargetErrorAtCoil, units=' mm', label='Target error at coil'),
            MetricSpecification(getter=self.getTargetXErrorAtCoil, units=' mm', label='Target X error at coil', doShowByDefault=False),
            MetricSpecification(getter=self.getTargetYErrorAtCoil, units=' mm', label='Target Y error at coil', doShowByDefault=False),
            MetricSpecification(getter=self.getDepthOffsetError, units=' mm', label='Depth offset error'),
            MetricSpecification(getter=self.getDepthAngleError, units='°', label='Depth angle error'),
            MetricSpecification(getter=self.getDepthTargetXAngleError, units='°', label='Depth target X angle error', doShowByDefault=False),
            MetricSpecification(getter=self.getDepthTargetYAngleError, units='°', label='Depth target Y angle error', doShowByDefault=False),
            MetricSpecification(getter=self.getDepthCoilXAngleError, units='°', label='Depth coil X angle error', doShowByDefault=False),
            MetricSpecification(getter=self.getDepthCoilYAngleError, units='°', label='Depth coil Y angle error', doShowByDefault=False),
            MetricSpecification(getter=self.getNormalCoilXAngleError, units='°', label='Normal coil X angle error', doShowByDefault=False),
            MetricSpecification(getter=self.getNormalCoilYAngleError, units='°', label='Normal coil Y angle error', doShowByDefault=False),
            MetricSpecification(getter=self.getHorizAngleError, units='°', label='Horiz angle error'),
            MetricSpecification(getter=self.getAngleFromMidline, units='°', label='Angle from midline'),
            MetricSpecification(getter=self.getAngleFromNormal, units='°', label='Angle from normal'),
            MetricSpecification(getter=self.getCoilPosX, units=' mm', label='Coil X position', doShowByDefault=False),
            MetricSpecification(getter=self.getCoilPosY, units=' mm', label='Coil Y position', doShowByDefault=False),
            MetricSpecification(getter=self.getCoilPosZ, units=' mm', label='Coil Z position', doShowByDefault=False),
        ])

    def _cacheWrap(self, fn: tp.Callable[..., T]) -> T:
        """
        Note: this assumes no other args or kwargs are needed for fn, since their values are not
        included in the cache key
        """
        # noinspection PyUnresolvedReferences
        key = fn.cacheKey

        if key in self._cachedValues:
            return self._cachedValues[key]
        else:
            val = fn(doUseCache=False)
            self._cachedValues[key] = val
            return val

    def getValueForMetric(self, label: str):
        whichMetric = None
        for metric in self._supportedMetrics:
            if metric.label == label:
                whichMetric = metric
        if whichMetric is None:
            raise KeyError(f'No metric with label {label} found')

        return whichMetric.getter()

    @property
    def supportedMetrics(self):
        return self._supportedMetrics

    @property
    def session(self):
        return self._session

    @property
    def sample(self):
        return self._sample

    @sample.setter
    def sample(self, newSample: Sample | None):
        if self._sample is newSample:
            return
        logger.debug('Sample changed')
        if self._sample is not None:
            self._sample.sigItemChanged.disconnect(self._onSampleChanged)
        self._sample = newSample
        if self._sample is not None:
            self._sample.sigItemChanged.connect(self._onSampleChanged)
        self._clearCachedValues()

    def _onTargetsChanged(self, targetKeys: list[str], targetAttrs: tp.Optional[list[str]] = None):
        if self._sample is not None and self._sample.targetKey in targetKeys and \
                (targetAttrs is None or len(set(targetAttrs) - {'isVisible'}) > 0):
            self._clearCachedValues(exceptKeys=[
                self.getAngleFromMidline.cacheKey,
                self.getAngleFromNormal.cacheKey,
                self.getCoilPosX.cacheKey,
                self.getCoilPosY.cacheKey,
                self.getCoilPosZ.cacheKey,
            ])

    def _onSampleChanged(self, sampleKey: str, sampleAttrs: tp.Optional[list[str]]):
        """
        Called when sample.sigItemChanged is emitted, not when `self.sample = ...` setter is called.
        """
        if sampleAttrs is None or len(set(sampleAttrs) - {'isVisible', 'isSelected', 'timestamp'}) > 0:
            logger.debug(f'PoseMetricCalculator onSampleChanged')
            if sampleAttrs is None or 'targetKey' in sampleAttrs:
                self._clearCachedValues()
            else:
                self._clearCachedValues(exceptKeys=[
                    self.getTargetCoilToScalpDist.cacheKey,
                    self.getTargetCoilToCortexDist.cacheKey,
                ])

    def _onFiducialsChanged(self, fidKeys: list[str], attribs: tp.Optional[list[str]] = None):
        if attribs is None or 'plannedCoord' in attribs:
            self._onPlannedFiducialsChanged()

    def _onPlannedFiducialsChanged(self):
        self._clearCachedValues(includingKeys=[self.getAngleFromMidline.cacheKey])

    def _clearCachedValues(self, includingKeys: tp.Optional[list[str]] = None, exceptKeys: tp.Optional[list[str]] = None):
        if includingKeys is None and exceptKeys is None:
            self._cachedValues = dict()
        else:
            if includingKeys is None:
                includingKeys = list(self._cachedValues.keys())
            if exceptKeys is None:
                exceptKeys = []
            self._cachedValues = {key: val for key, val in self._cachedValues.items()
                                  if (key in exceptKeys or key not in includingKeys)}
        self.sigCacheReset.emit()

    def _getTargetErrorAtDepth(self, depthFromTargetCoil: float, axis: tp.Optional[int] = None) -> float:

        target = self._session.targets[self._sample.targetKey]

        targetLinePts_targetCoilSpace = np.asarray([[0, 0, -depthFromTargetCoil], [0, 0, -depthFromTargetCoil + 10]])

        plane = Plane(point=targetLinePts_targetCoilSpace[0, :].squeeze(),
                      normal=np.diff(targetLinePts_targetCoilSpace, axis=0).squeeze())

        sampleLinePts_sampleCoilSpace = np.asarray([[0, 0, 0], [0, 0, 1]])
        sampleLinePts_targetCoilSpace = applyTransform([self._sample.coilToMRITransf, invertTransform(target.coilToMRITransf)], sampleLinePts_sampleCoilSpace, doCheck=False)
        line = Line(sampleLinePts_targetCoilSpace[0, :].squeeze(), np.diff(sampleLinePts_targetCoilSpace, axis=0).squeeze())

        try:
            samplePtOnPlane = plane.intersect_line(line)
        except ValueError:
            # sample axis is parallel to plane
            return np.nan

        if axis is None:
            dist = np.linalg.norm(targetLinePts_targetCoilSpace[0, :].squeeze() - samplePtOnPlane)
        else:
            dist = targetLinePts_targetCoilSpace[0, axis].squeeze() - samplePtOnPlane[axis]  # note that this is signed

        return dist

    def getTargetCoilToScalpDist(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(self.getTargetCoilToScalpDist)

        if self._sample is None or self._sample.targetKey is None:
            return np.nan

        skinSurf = self._session.headModel.skinSurf
        if skinSurf is None:
            return np.nan

        target = self._session.targets[self._sample.targetKey]

        return self._getCoilToSurfDist(coilToMRITransf=target.coilToMRITransf,
                                       surf=skinSurf)

    getTargetCoilToScalpDist.cacheKey = 'targetCoilToScalpDist'

    def getTargetCoilToCortexDist(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(self.getTargetCoilToCortexDist)

        if self._sample is None or self._sample.targetKey is None:
            return np.nan

        gmSurf = self._session.headModel.gmSurf
        if gmSurf is None:
            return np.nan

        target = self._session.targets[self._sample.targetKey]

        return self._getCoilToSurfDist(coilToMRITransf=target.coilToMRITransf,
                                       surf=gmSurf)

    getTargetCoilToCortexDist.cacheKey = 'targetCoilToCortexDist'

    def getSampleCoilToScalpDist(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(self.getSampleCoilToScalpDist)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        skinSurf = self._session.headModel.skinSurf
        if skinSurf is None:
            return np.nan

        return self._getCoilToSurfDist(coilToMRITransf=self._sample.coilToMRITransf,
                                       surf=skinSurf)

    getSampleCoilToScalpDist.cacheKey = 'sampleCoilToScalpDist'

    def getSampleCoilToCortexDist(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(self.getSampleCoilToCortexDist)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        gmSurf = self._session.headModel.gmSurf
        if gmSurf is None:
            return np.nan

        return self._getCoilToSurfDist(coilToMRITransf=self._sample.coilToMRITransf,
                                       surf=gmSurf)

    getSampleCoilToCortexDist.cacheKey = 'sampleCoilToCortexDist'

    def _getCoilToSurfDist(self, coilToMRITransf: np.ndarray, surf: pv.PolyData) -> float:
        coilOrigin_MRI = applyTransform(coilToMRITransf, np.zeros((3,)), doCheck=False)

        # TODO: maybe use additional constraint to find distance within a small range along coil depth axis
        # (e.g. a small sphere sliding down along the depth axis until reaching cortex)
        # Currently, this may find a closest point at a very oblique angle from coil center if coil is tilted

        closestPtIndex = find_closest_point(surf, coilOrigin_MRI)
        closestPt = surf.points[closestPtIndex, :]

        if False:
            # unsigned distance
            return np.linalg.norm(closestPt - coilOrigin_MRI)
        else:
            # signed distance, where coil -Z axis pointing down to surface is positive offset
            closestPt_coilSpace = applyTransform(invertTransform(coilToMRITransf), closestPt, doCheck=False)
            return -1*closestPt_coilSpace[2]

    def getTargetErrorInBrain(self, doUseCache: bool = True) -> float:
        """
        Distance from target in brain to sample after projecting down to plane of cortical target.

        Note: this determines depth of the target in the brain by finding closest point to the gm surf,
        so that we don't need to assume that target.targetCoord is already at the cortical surface.
        """
        if doUseCache:
            return self._cacheWrap(fn=self.getTargetErrorInBrain)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        coilTargetToCortexDist = self.getTargetCoilToCortexDist()

        if np.isnan(coilTargetToCortexDist):
            return np.nan

        return self._getTargetErrorAtDepth(depthFromTargetCoil=coilTargetToCortexDist)

    getTargetErrorInBrain.cacheKey = 'targetErrorInBrain'

    def getTargetErrorAtCoil(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(self.getTargetErrorAtCoil)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        if self._sample.targetKey is None:
            return np.nan

        return self._getTargetErrorAtDepth(depthFromTargetCoil=0)

    getTargetErrorAtCoil.cacheKey = 'targetErrorAtCoil'

    def getTargetXErrorAtCoil(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(self.getTargetXErrorAtCoil)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        if self._sample.targetKey is None:
            return np.nan

        return self._getTargetErrorAtDepth(depthFromTargetCoil=0, axis=0)  # note that this is relative to the target X axis, not sample X axis

    getTargetXErrorAtCoil.cacheKey = 'targetXErrorAtCoil'

    def getTargetYErrorAtCoil(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(self.getTargetYErrorAtCoil)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        if self._sample.targetKey is None:
            return np.nan

        return self._getTargetErrorAtDepth(depthFromTargetCoil=0, axis=1)  # note that this is relative to the target X axis, not sample X axis

    getTargetYErrorAtCoil.cacheKey = 'targetYErrorAtCoil'

    def getDepthOffsetError(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getDepthOffsetError)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        if self._sample.targetKey is None:
            return np.nan

        target = self._session.targets[self._sample.targetKey]

        sampleCoilPt_sampleCoilSpace = np.asarray([0, 0, 0])
        sampleCoilPt_targetCoilSpace = applyTransform(
            [self._sample.coilToMRITransf, invertTransform(target.coilToMRITransf)],
            sampleCoilPt_sampleCoilSpace, doCheck=False)

        offset = sampleCoilPt_targetCoilSpace[2]

        return offset

    getDepthOffsetError.cacheKey = 'depthOffsetError'

    def getDepthAngleError(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getDepthAngleError)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        if self._sample.targetKey is None:
            return np.nan

        target = self._session.targets[self._sample.targetKey]

        targetLinePts_targetCoilSpace = np.asarray([[0, 0, 0], [0, 0, 1]])
        targetLinePts_MRISpace = applyTransform(target.coilToMRITransf, targetLinePts_targetCoilSpace, doCheck=False)
        targetVector = Vector(np.diff(targetLinePts_MRISpace, axis=0).squeeze())

        sampleLinePts_sampleCoilSpace = np.asarray([[0, 0, 0], [0, 0, 1]])
        sampleLinePts_MRISpace = applyTransform(self._sample.coilToMRITransf, sampleLinePts_sampleCoilSpace, doCheck=False)
        sampleVector = Vector(np.diff(sampleLinePts_MRISpace, axis=0).squeeze())

        angle = targetVector.angle_between(sampleVector)

        return np.rad2deg(angle)

    getDepthAngleError.cacheKey = 'depthAngleError'

    def _getDepthComponentAngleError(self, iDim: int, relTo: str):
        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        if self._sample.targetKey is None:
            return np.nan

        target = self._session.targets[self._sample.targetKey]

        match relTo:
            case 'target':
                targetLinePts_targetCoilSpace = np.asarray([[0, 0, 0], [0, 0, 1]])

                sampleLinePts_sampleCoilSpace = np.asarray([[0, 0, 0], [0, 0, 1]])
                sampleLinePts_targetCoilSpace = applyTransform([self._sample.coilToMRITransf,
                                                                invertTransform(target.coilToMRITransf)],
                                                                sampleLinePts_sampleCoilSpace, doCheck=False)

                targetVector = Vector(np.diff(targetLinePts_targetCoilSpace[:, [iDim, 2]], axis=0).squeeze())
                sampleVector = Vector(np.diff(sampleLinePts_targetCoilSpace[:, [iDim, 2]], axis=0).squeeze())

            case 'coil':
                targetLinePts_targetCoilSpace = np.asarray([[0, 0, 0], [0, 0, 1]])
                targetLinePts_sampleCoilSpace = applyTransform([target.coilToMRITransf,
                                                               invertTransform(self._sample.coilToMRITransf)],
                                                               targetLinePts_targetCoilSpace, doCheck=False)

                sampleLinePts_sampleCoilSpace = np.asarray([[0, 0, 0], [0, 0, 1]])

                targetVector = Vector(np.diff(targetLinePts_sampleCoilSpace[:, [iDim, 2]], axis=0).squeeze())
                sampleVector = Vector(np.diff(sampleLinePts_sampleCoilSpace[:, [iDim, 2]], axis=0).squeeze())

            case _:
                raise NotImplementedError

        angle = targetVector.angle_signed(sampleVector)

        return np.rad2deg(angle)

    def getDepthTargetXAngleError(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getDepthTargetXAngleError)

        return self._getDepthComponentAngleError(iDim=0, relTo='target')

    getDepthTargetXAngleError.cacheKey = 'depthTargetXAngleError'

    def getDepthTargetYAngleError(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getDepthTargetYAngleError)

        return self._getDepthComponentAngleError(iDim=1, relTo='target')

    getDepthTargetYAngleError.cacheKey = 'depthTargetYAngleError'

    def getDepthCoilXAngleError(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getDepthCoilXAngleError)

        return self._getDepthComponentAngleError(iDim=0, relTo='coil')

    getDepthCoilXAngleError.cacheKey = 'depthCoilXAngleError'

    def getDepthCoilYAngleError(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getDepthCoilYAngleError)

        return self._getDepthComponentAngleError(iDim=1, relTo='coil')

    getDepthCoilYAngleError.cacheKey = 'depthCoilYAngleError'

    def getHorizAngleError(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getHorizAngleError)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        if self._sample.targetKey is None:
            return np.nan

        target = self._session.targets[self._sample.targetKey]

        # project point representing sample coil handle onto horizontal plane of target coil, then determine angle between handles

        sampleLinePts_sampleCoilSpace = np.asarray([[0, 0, 0], [0, -1, 0]])
        sampleLinePts_targetCoilSpace = applyTransform(
            [self._sample.coilToMRITransf, invertTransform(target.coilToMRITransf)],
            sampleLinePts_sampleCoilSpace, doCheck=False)

        targetHandleVector2D_targetCoilSpace = Vector([0, -1])
        sampleHandleVector2D_targetCoilSpace = Vector(np.diff(sampleLinePts_targetCoilSpace[:, 0:2], axis=0).squeeze())

        angle = targetHandleVector2D_targetCoilSpace.angle_signed(sampleHandleVector2D_targetCoilSpace)

        return np.rad2deg(angle)

    getHorizAngleError.cacheKey = 'horizAngleError'

    def getAngleFromMidline(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getAngleFromMidline)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        return calculateAngleFromMidlineFromCoilToMRITransf(self.session, self._sample.coilToMRITransf)

    getAngleFromMidline.cacheKey = 'angleFromMidline'

    def getAngleFromNormal(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getAngleFromNormal)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        coilToCortexDist = self.getSampleCoilToCortexDist()
        if np.isnan(coilToCortexDist):
            return np.nan

        if False:
            pt_gm = self.getClosestPointToCoilOnGM()
            closestPt_skin = self.getClosestPointToCoilOnSkin()
        else:
            pt_gm = applyTransform(self._sample.coilToMRITransf, np.asarray([0, 0, -coilToCortexDist]), doCheck=False)
            closestPt_skin = self.getClosestPointToSampleCortexDepthOnSkin()

        if closestPt_skin is None or pt_gm is None:
            return np.nan

        # find ideal normal by defining line through these two points
        idealNormal = Vector(closestPt_skin - pt_gm)
        actualNormal = Vector(np.diff(applyTransform(self._sample.coilToMRITransf,
                                                     np.asarray([[0, 0, -1], [0, 0, 0,]]),
                                                     doCheck=False), axis=0).squeeze())

        angle = idealNormal.angle_between(actualNormal)

        return np.rad2deg(angle)

    getAngleFromNormal.cacheKey = 'angleFromNormal'

    def getCoilPosX(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getCoilPosX)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        return self._sample.coilToMRITransf[0, 3]

    getCoilPosX.cacheKey = 'coilPosX'

    def getCoilPosY(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getCoilPosY)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        return self._sample.coilToMRITransf[1, 3]

    getCoilPosY.cacheKey = 'coilPosY'

    def getCoilPosZ(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getCoilPosZ)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        return self._sample.coilToMRITransf[2, 3]

    getCoilPosZ.cacheKey = 'coilPosZ'

    def getClosestPointToCoilOnSkin(self, doUseCache: bool = True) -> tp.Optional[np.ndarray]:
        if doUseCache:
            return self._cacheWrap(fn=self.getClosestPointToCoilOnSkin)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return None

        return getClosestPointToPointOnMesh(session=self._session,
                                            whichMesh='skinSurf',
                                            point_MRISpace=applyTransform(self._sample.coilToMRITransf, np.zeros((3,)), doCheck=False))

    getClosestPointToCoilOnSkin.cacheKey = 'closestPointToCoilOnSkin'

    def getClosestPointToCoilOnGM(self, doUseCache: bool = True) -> tp.Optional[np.ndarray]:
        if doUseCache:
            return self._cacheWrap(fn=self.getClosestPointToCoilOnGM)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return None

        return getClosestPointToPointOnMesh(session=self._session,
                                            whichMesh='gmSurf',
                                            point_MRISpace=applyTransform(self._sample.coilToMRITransf, np.zeros((3,)), doCheck=False))

    getClosestPointToCoilOnGM.cacheKey = 'closestPointToCoilOnGM'

    def getClosestPointToSampleCortexDepthOnSkin(self, doUseCache: bool = True) -> tp.Optional[np.ndarray]:
        if doUseCache:
            return self._cacheWrap(fn=self.getClosestPointToSampleCortexDepthOnSkin)

        if self._sample is None or self._sample.coilToMRITransf is None:
            return None

        coilToCortexDist = self.getSampleCoilToCortexDist()
        if np.isnan(coilToCortexDist):
            return None

        point_MRISpace = applyTransform(self._sample.coilToMRITransf, np.asarray([0, 0, -coilToCortexDist]), doCheck=False)

        return getClosestPointToPointOnMesh(session=self._session,
                                            whichMesh='skinSurf',
                                            point_MRISpace=point_MRISpace)

    getClosestPointToSampleCortexDepthOnSkin.cacheKey = 'closestPointToSampleCortexDepthOnSkin'

    def _getNormalCoilAngleError(self, iDim: int) -> float:

        if self._sample is None or self._sample.coilToMRITransf is None:
            return np.nan

        coilToCortexDist = self.getSampleCoilToCortexDist()
        if np.isnan(coilToCortexDist):
            return np.nan

        pt_gm = applyTransform(self._sample.coilToMRITransf, np.asarray([0, 0, -coilToCortexDist]), doCheck=False)
        closestPt_skin = self.getClosestPointToSampleCortexDepthOnSkin()

        if closestPt_skin is None:
            return np.nan

        # find ideal normal by defining line through these two points
        idealNormalPts_mriSpace = np.vstack([pt_gm, closestPt_skin])
        idealNormalPts_coilSpace = applyTransform(invertTransform(self._sample.coilToMRITransf), idealNormalPts_mriSpace, doCheck=False)
        idealNormal = Vector(np.diff(idealNormalPts_coilSpace[:, [iDim, 2]], axis=0).squeeze())
        actualNormal = Vector([0, 1])

        angle = idealNormal.angle_signed(actualNormal)  # TODO: check sign

        return np.rad2deg(angle)

    def getNormalCoilXAngleError(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getNormalCoilXAngleError)

        return self._getNormalCoilAngleError(iDim=0)

    getNormalCoilXAngleError.cacheKey = 'normalCoilXAngleError'

    def getNormalCoilYAngleError(self, doUseCache: bool = True) -> float:
        if doUseCache:
            return self._cacheWrap(fn=self.getNormalCoilYAngleError)

        return self._getNormalCoilAngleError(iDim=1)

    getNormalCoilYAngleError.cacheKey = 'normalCoilYAngleError'



from __future__ import annotations

import shutil

import attrs
from datetime import datetime
import nibabel as nib
import json
import logging
import numpy as np
import os
import pandas as pd
import pyvista as pv
import tempfile
import typing as tp
from typing import ClassVar

from NaviNIBS.util.attrs import attrsAsDict
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.Transforms import invertTransform
from NaviNIBS.util.numpy import array_equalish


logger = logging.getLogger(__name__)


@attrs.define()
class MRI:
    _filepath: tp.Optional[str] = None

    _clim2DMin: float | None = None
    _clim2DMax: float | None = None
    _clim3DMin: float | None = None
    _clim3DMax: float | None = None

    _data: tp.Optional[nib.Nifti1Image] = attrs.field(init=False, default=None)
    _dataAsUniformGrid: tp.Optional[pv.ImageData] = attrs.field(init=False, default=None)
    _inverseAffine: np.ndarray | None = attrs.field(init=False, default=None)
    _autoClim2DMin: float | None = attrs.field(init=False, default=None)
    _autoClim2DMax: float | None = attrs.field(init=False, default=None)
    _autoClim3DMin: float | None = attrs.field(init=False, default=None)
    _autoClim3DMax: float | None = attrs.field(init=False, default=None)

    sigFilepathChanged: Signal = attrs.field(init=False, factory=Signal)
    sigDataChanged: Signal = attrs.field(init=False, factory=Signal)
    sigManualClimChanged: Signal[str] = attrs.field(init=False, factory=Signal)
    """
    Includes label of which clim changed ('2D' or '3D')
    """
    sigClimChanged: Signal[str] = attrs.field(init=False, factory=Signal)
    """
    Includes label of which clim changed ('2D' or '3D')
    """

    def __attrs_post_init__(self):
        self.sigFilepathChanged.connect(self._onFilepathChanged)
        self.validateFilepath(self._filepath)

    def loadCache(self):
        if not self.isSet:
            logger.warning('Load data requested, but no filepath set. Returning.')
            return

        logger.info('Loading image into cache from {}'.format(self.filepath))
        self._data = nib.load(self.filepath)

        if True:
            # create pyvista data object

            if pv.__version__ <= '0.39.1':
                self._dataAsUniformGrid = pv.UniformGrid(
                    dims=self._data.shape)
            else:
                self._dataAsUniformGrid = pv.ImageData(
                    dimensions=self._data.shape)

            self._dataAsUniformGrid.point_data['MRI'] = np.asanyarray(self.data.dataobj).ravel(order='F')

        if True:
            # cache inverse of affine transform
            self._inverseAffine = invertTransform(self._data.affine)

        # clear cached auto clim values
        for dim in ('2D', '3D'):
            for minOrMax in ('Min', 'Max'):
                propKey = f'_autoClim{dim}{minOrMax}'
                setattr(self, propKey, None)

        self.sigDataChanged.emit()

        if self._clim2DMin is None or self._clim2DMax is None:
            self.sigClimChanged.emit('2D')

        if self._clim3DMin is None or self._clim3DMax is None:
            self.sigClimChanged.emit('3D')

    def clearCache(self):
        if self._data is None:
            return
        self._data = None
        self._dataAsUniformGrid = None
        self._inverseAffine = None
        # clear cached auto clim values
        for dim in ('2D', '3D'):
            for minOrMax in ('Min', 'Max'):
                propKey = f'_autoClim{dim}{minOrMax}'
                setattr(self, propKey, None)

    def _onFilepathChanged(self):
        with self.sigDataChanged.blocked():
            self.clearCache()
        self.sigDataChanged.emit()

        if self._clim2DMin is None or self._clim2DMax is None:
            self.sigClimChanged.emit('2D')

        if self._clim3DMin is None or self._clim3DMax is None:
            self.sigClimChanged.emit('3D')

    @property
    def filepath(self):
        return self._filepath

    @filepath.setter
    def filepath(self, newPath: str):
        if self._filepath == newPath:
            return
        self.validateFilepath(newPath)
        logger.info(f'Changing MRI filepath from {self._filepath} to {newPath}')
        self._filepath = newPath
        self.sigFilepathChanged.emit()
        # TODO: here or with slots connected to sigDataChanged, make sure any cached MRI data or metadata is cleared/reloaded

    @property
    def isSet(self):
        return self._filepath is not None

    def loadCacheIfNeeded(self):
        if self.isSet and self._data is None:
            # data was not previously loaded, but it is available. Load now.
            self.loadCache()

    @property
    def data(self):
        self.loadCacheIfNeeded()
        return self._data

    @property
    def dataAsUniformGrid(self):
        """
        Note: this data is in the original image space (coordinate indices), without any transformations applied.
        """
        self.loadCacheIfNeeded()
        return self._dataAsUniformGrid

    @property
    def dataToScannerTransf(self) -> np.ndarray | None:
        """
        Transform from image array (coordinate indices) to native MRI space.

        See https://nipy.org/nibabel/coordinate_systems.html for more info.
        """
        self.loadCacheIfNeeded()
        if self._data is None:
            return None
        else:
            return self._data.affine

    @property
    def scannerToDataTransf(self) -> np.ndarray | None:
        """
        Inverse of ``dataToScannerTransf``
        """
        self.loadCacheIfNeeded()
        return self._inverseAffine

    @property
    def clim2DMin(self):
        """
        Manually-set minimum of color axis for 2D plots, or None to autoset.
        """
        return self._clim2DMin

    @clim2DMin.setter
    def clim2DMin(self, newVal: float | None):
        if self._clim2DMin == newVal:
            return
        logger.info(f'2D clim min changing from {self._clim2DMin} to {newVal}')
        self._clim2DMin = newVal
        self.sigClimChanged.emit('2D')
        self.sigManualClimChanged.emit('2D')

    @property
    def clim2DMax(self):
        """
        Manually-set maximum of color axis for 2D plots, or None to autoset.
        """
        return self._clim2DMax

    @clim2DMax.setter
    def clim2DMax(self, newVal: float | None):
        if self._clim2DMax == newVal:
            return
        logger.info(f'2D clim max changing from {self._clim2DMax} to {newVal}')
        self._clim2DMax = newVal
        self.sigClimChanged.emit('2D')
        self.sigManualClimChanged.emit('2D')

    @property
    def clim2D(self) -> tuple[float | None, float | None]:
        """
        Merged manually-set and auto-set clims depending on what is available.

        Only returns tuple of Nones if data is not available.
        """
        return (
            self.clim2DMin if self.clim2DMin is not None else self.autoClim2DMin,
            self.clim2DMax if self.clim2DMax is not None else self.autoClim2DMax
        )

    @property
    def clim3DMin(self):
        """
        Manually-set minimum of color axis for 3D plots, or None to autoset.
        """

        return self._clim3DMin

    @clim3DMin.setter
    def clim3DMin(self, newVal: float | None):
        if self._clim3DMin == newVal:
            return
        logger.info(f'3D clim min changing from {self._clim3DMin} to {newVal}')
        self._clim3DMin = newVal
        self.sigClimChanged.emit('3D')
        self.sigManualClimChanged.emit('3D')

    @property
    def clim3DMax(self):
        """
        Manually-set maximum of color axis for 3D plots, or None to autoset.
        """

        return self._clim3DMax

    @clim3DMax.setter
    def clim3DMax(self, newVal: float | None):
        if self._clim3DMax == newVal:
            return
        logger.info(f'3D clim max changing from {self._clim3DMax} to {newVal}')
        self._clim3DMax = newVal
        self.sigClimChanged.emit('3D')
        self.sigManualClimChanged.emit('3D')

    @property
    def clim3D(self):
        """
        Merged manually-set and auto-set clims depending on what is available.

        Only returns tuple of Nones if data is not available.
        """
        return (
            self.clim3DMin if self.clim3DMin is not None else self.autoClim3DMin,
            self.clim3DMax if self.clim3DMax is not None else self.autoClim3DMax
        )

    def _calculateAutoClim(self, dim: str, minOrMax: str):
        propKey = f'_autoClim{dim}{minOrMax}'

        if self._data is None:
            raise ValueError('No data loaded, cannot calculate auto clim')
        else:
            data = self._data.get_fdata()

            lowerThreshold = np.percentile(data, 90) / 10
            #logger.debug(f'Lower threshold for auto clim calculation: {lowerThreshold}')

            if minOrMax == 'Min':
                if dim == '2D':
                    pct = 1
                else:
                    pct = 20
            else:
                if dim == '2D':
                    pct = 95
                else:
                    pct = 80

            val = np.nanpercentile(np.where(data > lowerThreshold, data, np.nan), pct)

            logger.debug(f'Calculated auto clim {dim}{minOrMax}: {val}')

            setattr(self, propKey, val)

            self.sigClimChanged.emit(dim)

    def _getAutoClim(self, dim: str, minOrMax: str) -> float | None:
        self.loadCacheIfNeeded()
        if self._data is None:
            return None
        else:
            propKey = f'_autoClim{dim}{minOrMax}'
            if getattr(self, propKey) is None:
                self._calculateAutoClim(dim=dim, minOrMax=minOrMax)
            return getattr(self, propKey)

    @property
    def autoClim2DMin(self):
        return self._getAutoClim(dim='2D', minOrMax='Min')

    @property
    def autoClim2DMax(self):
        return self._getAutoClim(dim='2D', minOrMax='Max')

    @property
    def autoClim3DMin(self):
        return self._getAutoClim(dim='3D', minOrMax='Min')

    @property
    def autoClim3DMax(self):
        return self._getAutoClim(dim='3D', minOrMax='Max')

    def asDict(self, filepathRelTo: str) -> tp.Dict[str, tp.Any]:
        d = attrsAsDict(self)
        if 'filepath' in d:
            d['filepath'] = os.path.relpath(d['filepath'], filepathRelTo)

        return d

    @classmethod
    def fromDict(cls, d: tp.Dict[str, tp.Any], filepathRelTo: str) -> MRI:
        # TODO: validate against schema
        if 'filepath' in d:
            d['filepath'] = os.path.join(filepathRelTo, d['filepath'])
            cls.validateFilepath(d['filepath'])
        return cls(**d)

    @classmethod
    def validateFilepath(cls, filepath: tp.Optional[str]) -> None:
        if filepath is None:
            return
        assert filepath.endswith('.nii') or filepath.endswith('.nii.gz')
        assert os.path.exists(filepath), 'File not found at {}'.format(filepath)




from __future__ import annotations

from abc import ABC
import attrs
import importlib
import json
import logging
import os
import sys
import typing as tp
from typing import TYPE_CHECKING

from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.attrs import attrsAsDict
from NaviNIBS.util.json import jsonPrettyDumps

from NaviNIBS.Navigator.Model.GenericCollection import GenericCollection, GenericCollectionDictItem

if TYPE_CHECKING:
    from NaviNIBS.Navigator.Model.Session import Session
    from NaviNIBS.Navigator.GUI.NavigatorGUI import NavigatorGUI
    from NaviNIBS.Navigator.GUI.ViewPanels import MainViewPanel
    from NaviNIBS.Navigator.GUI.ViewPanels.NavigatePanel.NavigationView import NavigationView
    from NaviNIBS.Navigator.GUI.ViewPanels.NavigatePanel.ViewLayers import ViewLayer

logger = logging.getLogger(__name__)


installPath = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', '..')


ACE = tp.TypeVar('ACE')  # type (e.g. MainViewPanel) referenced by each addon class element


@attrs.define
class AddonSessionConfig(ABC):
    """
    Base class to define an addon's configuration parameters that will be saved in session config
    """

    sigConfigAboutToChange: Signal = attrs.field(init=False, factory=lambda: Signal((tp.Optional[list[str]],)))
    # should be emitted whenever a field that is included in serialized asDict() is about to change
    sigConfigChanged: Signal = attrs.field(init=False, factory=lambda: Signal((tp.Optional[list[str]],)))
    # should be emitted whenever a field that is included in serialized asDict() has changed

    def __attrs_post_init__(self):
        pass

    def asDict(self) -> dict[str, tp.Any]:
        # can be overridden by subclass
        return attrsAsDict(self)

    @classmethod
    def fromDict(cls, d: dict[str, tp.Any]):
        # can be overridden by subclass
        return cls(**d)


@attrs.define
class AddonClassElement(tp.Generic[ACE]):
    _ClassName: str
    _importModule: str
    _key: str = None  # if not specified, will be same as ClassName
    isBuiltIn: bool = False
    _Class: tp.Optional[tp.Type[ACE]] = attrs.field(init=False, default=None)

    def __attrs_post_init__(self):
        if self._importModule is not None:
            assert self._ClassName is not None  # need both importModule and ClassName to fully specify what to import

            if False:
                # disabled for now since it seems to cause some import errors / long delays
                # TODO: troubleshoot problems and set this to be optionally enabled by default
                logger.debug('Verifying specified importModule is in path')  # TODO: debug, delete log statements
                if importlib.util.find_spec(self._importModule) is None:
                    raise ModuleNotFoundError('Module "%s" not found in python path' % (self._importModule,))
                logger.debug('Verified')
        else:
            assert self._Class is not None
            if self._ClassName is None:
                self._ClassName = self._Class.__name__
            else:
                assert self._ClassName == self._Class.__name__

    @property
    def key(self):
        if self._key is None:
            return self._ClassName
        else:
            return self._key

    @property
    def ClassName(self):
        return self._ClassName

    @property
    def importModule(self):
        return self._importModule

    @property
    def Class(self) -> tp.Type[ACE]:
        if self._Class is None:
            self.reload()
            assert self._Class is not None
        return self._Class

    def reload(self):
        assert self._importModule is not None
        logger.debug('Importing module %s' % (self._importModule,))
        module = importlib.import_module(self._importModule)
        self._Class = getattr(module, self._ClassName)

    def asDict(self) -> tp.Dict[str, tp.Any]:
        return attrsAsDict(self)

    @classmethod
    def fromDict(cls, d: tp.Dict[str, tp.Any]):
        return cls(**d)


@attrs.define
class AddonExtra(ABC):
    """
    Base class to define optional custom code for an addon.

    Use this when an addon needs to provide custom code not part of a main view panel, navigation view layer, etc. This may be used for code-only addon pieces (such as providing a data output stream), or for patching existing GUI functionality (using the reference to the root NavigatorGUI instance).

    Any AddonCode class elements defined in an addon configuration will be instantiated and passed references to the root NavigatorGUI instance and the session model when the addon is loaded.
    """
    _navigatorGUI: NavigatorGUI
    _session: Session = attrs.field(repr=False)

    def __attrs_post_init__(self):
        pass


@attrs.define
class Addon(GenericCollectionDictItem[str]):
    _addonInstallPath: str
    _addonVersion: str | None = None
    _MainViewPanels: tp.Dict[str, AddonClassElement[MainViewPanel]] = attrs.field(factory=dict)
    _NavigationViews: tp.Dict[str, AddonClassElement[NavigationView]] = attrs.field(factory=dict)
    _NavigationViewLayers: tp.Dict[str, AddonClassElement[ViewLayer]] = attrs.field(factory=dict)
    _Extras: tp.Dict[str, AddonClassElement[AddonExtra]] = attrs.field(factory=dict)
    _SessionAttrs: tp.Dict[str, AddonClassElement[AddonSessionConfig]] = attrs.field(factory=dict)

    _sessionAttrs: dict[str, AddonSessionConfig] = attrs.field(factory=dict)
    _extras: dict[str, AddonExtra] = attrs.field(init=False, factory=dict)
    _isActive: bool = True

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        logger.info(f'Initializing addon {self.key} (version {self._addonVersion if self._addonVersion is not None else "not set"})')

        for key, SessionAttr in self._SessionAttrs.items():
            if key not in self._sessionAttrs:
                self._sessionAttrs[key] = SessionAttr.Class()

        for key, sessionAttr in self._sessionAttrs.items():
            assert isinstance(sessionAttr, self._SessionAttrs[key].Class)
            sessionAttr.sigConfigAboutToChange.connect(lambda *args: self.sigItemAboutToChange.emit(self.key, ['SessionAttrs.' + key]))
            sessionAttr.sigConfigChanged.connect(lambda *args: self.sigItemChanged.emit(self.key, ['SessionAttrs.' + key]))

    @property
    def MainViewPanels(self):
        return self._MainViewPanels

    @property
    def addonInstallPath(self):
        return self._addonInstallPath

    @property
    def SessionAttrs(self):
        return self._SessionAttrs

    @property
    def needsToInstantiateExtras(self):
        return len(self._Extras) > 0 and len(self._extras) == 0

    def instantiateExtras(self, navigatorGUI: NavigatorGUI, session: Session):
        if len(self._extras) > 0:
            raise RuntimeError('Addon extras already instantiated')

        for key, Extra in self._Extras.items():
            logger.info(f'Instantiating addon extra {key}')
            self._extras[key] = Extra.Class(navigatorGUI=navigatorGUI,
                                            session=session)

    def __getattr__(self, key: str):
        """
        Allow accessing session attributes with addon.attributeName
        """
        try:
            return self._sessionAttrs[key]
        except KeyError:
            raise AttributeError

    def asDict(self) -> tp.Dict[str, tp.Any]:
        predefinedAttrs = ['key', 'MainViewPanels', 'Extras', 'NavigationViews', 'NavigationViewLayers', 'SessionAttrs']  # these should all be defined in fixed addon_configuration.json file

        d = attrsAsDict(self, exclude=predefinedAttrs + ['sessionAttrs'])
        for key, sessionAttr in self._sessionAttrs.items():
            d[key] = sessionAttr.asDict()
            if len(d[key]) == 0:
                del d[key]

        if self._addonVersion is not None:
            d['addonVersion'] = self._addonVersion
        d['addonInstallPath'] = os.path.relpath(self.addonInstallPath, installPath)

        return d

    def getConfigFilename(self, unpackedSessionDir: str, filenamePrefix: str = '') -> str:
        return filenamePrefix + 'Addon_' + self.key + '.json'

    def writeConfig(self, unpackedSessionDir: str, filenamePrefix: str = '') -> str:
        configFilename_addon = self.getConfigFilename(
            unpackedSessionDir=unpackedSessionDir,
            filenamePrefix=filenamePrefix)
        toWrite = jsonPrettyDumps(self.asDict())
        with open(os.path.join(unpackedSessionDir, configFilename_addon), 'w') as f:
            f.write(toWrite)
        return configFilename_addon

    @classmethod
    def fromDict(cls, d: tp.Dict[str, tp.Any]) -> Addon:
        d = d.copy()
        # assume only path to addon_configuration and optionally SessionAttrs are specified

        assert 'addonInstallPath' in d

        # specified path in config is relative to root NaviNIBS installPath
        addonInstallPath = os.path.join(installPath, d['addonInstallPath'])
        del d['addonInstallPath']
        addonConfigPath = os.path.join(addonInstallPath, 'addon_configuration.json')

        if not os.path.exists(addonConfigPath):
            raise FileNotFoundError(f'Addon configuration not found at {addonConfigPath}')

        with open(addonConfigPath, 'r') as f:
            dc = json.load(f)

        if not os.path.split(addonInstallPath)[0] in sys.path:
            sys.path.append(os.path.split(addonInstallPath)[0])

        elementAttrs = ('MainViewPanels', 'Extras', 'NavigationViews', 'NavigationViewLayers', 'SessionAttrs')

        initKwargs = dict(
            addonInstallPath=addonInstallPath,
            key=dc['key']
        )

        if 'addonVersion' in dc:
            if 'addonVersion' in d:
                if d['addonVersion'] != dc['addonVersion']:
                    logger.warning(
                        f'Addon version mismatch: session config version {d["addonVersion"]} does not match addon configuration version {dc["addonVersion"]}. Using addon configuration version.')
                del d['addonVersion']
            initKwargs['addonVersion'] = dc['addonVersion']
        else:
            assert 'addonVersion' not in d, 'Addon version specified in session config but not in addon configuration'

        for elementAttr in elementAttrs:
            if elementAttr in dc:
                initKwargs[elementAttr] = dict()
                for elementKey, elementDict in dc[elementAttr].items():
                    initKwargs[elementAttr][elementKey] = AddonClassElement.fromDict(elementDict)

        if len(d) > 0:
            sessionAttrs = dict()
            for key, sessionAttrDict in d.items():
                try:
                    SessionAttr = initKwargs['SessionAttrs'][key].Class
                except KeyError:
                    raise KeyError(f'Addon session config included session attributes defined for {key}, but this was not found in addon_configuration.json')

                assert issubclass(SessionAttr, AddonSessionConfig)
                sessionAttrs[key] = SessionAttr.fromDict(sessionAttrDict)

            initKwargs['sessionAttrs'] = sessionAttrs
        else:
            pass  # no other addon-specific attrs defined in session config

        return cls(**initKwargs)


@attrs.define
class Addons(GenericCollection[str, Addon]):

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def asList(self, unpackedSessionDir: str,
               filenamePrefix: str = '',
               doWriteAddonConfigs: bool = True) -> tp.List[str]:
        addonSessionConfigFilenames = []
        for addonKey, addon in self.items():
            if doWriteAddonConfigs:
                configFilename_addon = addon.writeConfig(unpackedSessionDir=unpackedSessionDir,
                                                         filenamePrefix=filenamePrefix)
            else:
                configFilename_addon = addon.getConfigFilename(unpackedSessionDir=unpackedSessionDir,
                                                               filenamePrefix=filenamePrefix)
            addonSessionConfigFilenames.append(configFilename_addon)
        return addonSessionConfigFilenames

    @classmethod
    def fromList(cls, addonList: tp.List[str], unpackedSessionDir: str) -> Addons:
        addons = {}
        for addonSessionConfigFilename in addonList:
            with open(os.path.join(unpackedSessionDir, addonSessionConfigFilename), 'r') as f:
                addonDict = json.load(f)
            addon = Addon.fromDict(addonDict)
            addons[addon.key] = addon

        return cls(items=addons)






from __future__ import annotations

import attrs
import logging

import qtawesome as qta
from qtpy import QtGui
import typing as tp

from . import HotkeyTriggerSourceSettingsWidget
from .HotkeyTriggerSourceSettingsWidget import HotkeyTriggerSourceSettingsWidget

from .. import MainViewPanel
from .LSLTriggerSourceSettingsWidget import LSLTriggerSourceSettingsWidget
from .TriggerSourceSettingsWidget import TriggerSourceSettingsWidget
from NaviNIBS.Navigator.Model.Triggering import TriggerEvent
from NaviNIBS.Navigator.GUI.ViewPanels.MainViewPanelWithDockWidgets import MainViewPanelWithDockWidgets
from NaviNIBS.util.GUI.Icons import getIcon

logger = logging.getLogger(__name__)


@attrs.define
class TriggerSettingsPanel(MainViewPanelWithDockWidgets):
    _key: str = 'Trigger settings'
    _icon: QtGui.QIcon = attrs.field(init=False, factory=lambda: getIcon('mdi6.database-import'))

    _lslSettingsWidgets: list[LSLTriggerSourceSettingsWidget] = attrs.field(init=False, factory=list)
    _hotkeySettings: dict[str, HotkeyTriggerSourceSettingsWidget] = attrs.field(init=False, factory=dict)
    _hasInitializedTriggerSources: bool = attrs.field(init=False, default=False)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def _finishInitialization(self):
        super()._finishInitialization()

        if self.session is not None:
            self._onPanelInitializedAndSessionSet()

    def _onSessionSet(self):
        super()._onSessionSet()

        # initialize right away so that we will start listening to streams and hotkeys immediately
        assert not self._hasInitializedTriggerSources, 'Updating with new session not implemented'

        assert len(self._hotkeySettings) == 0, 'Updating with new session not implemented'
        for triggerSourceKey, triggerSource in self.session.triggerSources.items():
            match triggerSource.type:
                case 'LSLTriggerSource':
                    self._lslSettingsWidgets.append(LSLTriggerSourceSettingsWidget(
                        dockKey=self._key,
                        session=self.session,
                        triggerSourceKey=triggerSourceKey,
                    ))
                    self._wdgt.addDock(self._lslSettingsWidgets[-1].dock, position='left')
                case 'HotkeyTriggerSource':
                    self._hotkeySettings[triggerSourceKey] = HotkeyTriggerSourceSettingsWidget(
                        triggerSourceKey=triggerSourceKey,
                        title=f'{triggerSourceKey} hotkey settings',
                        dockKey=self._key,
                        session=self.session
                    )
                    self._wdgt.addDock(self._hotkeySettings[triggerSourceKey].dock, position='right')
                case _:
                    raise NotImplementedError(f'Unexpected trigger source type: {triggerSource.type}')

        self._hasInitializedTriggerSources = True

        if self._hasInitialized:
            self._onPanelInitializedAndSessionSet()

    def _onPanelInitializedAndSessionSet(self):
        pass

    def canBeEnabled(self) -> tuple[bool, str | None]:
        if self.session is None:
            return False, 'No session set'
        return True, None

    def _onTriggered(self, triggerEvt: TriggerEvent):
        pass  # TODO: show GUI indicator about time of last trigger(s)